import React, {useState, useEffect, useMemo, useRef}  from 'react';
import logo from './logo.svg';
import './App.css';

import { GraphQLClient } from 'graphql-request';

import createModels2 from './front-models'

let gql = new GraphQLClient("/graphql", {headers: localStorage.authToken ? {Authorization: 'Bearer '+localStorage.authToken} : {}})


//TODO: use graphql-tag to get types, or, at least detect query 
//for proper binding between model objects and result of gql query



const dataReader = async () => {
    const { PriceItem } = await createModels2(gql)
    let data            = await PriceItem.find({}, {limit: [10]})
    return {data, PriceItem }
}

const Th = p =>
<div className='Th' {...p} />

//<div className='Row' style={{position: 'relative',top}}>
const Row = ({top, children}) =>
<div className='Row' >
    {children}
</div>

const Cell = ({children, ...props}) =>
<div className='Cell' {...props}>
    {children && children.toString()}
</div>


const ModelListItem = "div"

const ModelList = ({models, selected, onChange, Item=ModelListItem}) => {
    return (
        <aside>
            {Object.entries(models).map(([name, model]) => {
                return (
                    <Item key={name} 
                                   selected={name === selected}
                                   className={name === selected && "selected"}
                                   onClick = {() => onChange(name)}
                                   >
                        {name}
                    </Item>
                )
            })}
        </aside>
    )
}

const Search = ({...props}) => 
<input placeholder="Search..." {...props} className="Search"/>

const Count = ({...props}) => <div className="Count" {...props} /> 

const searchQueryBuilder = (search, model) => {
    const queryRegexp = `/${search.trim().split(/\s+/).join('|')}/`
    return {$or: model.fields.filter(field => field.type.kind == 'SCALAR').map(field => ({[field.name]: queryRegexp}))}
}

const GridHeaderItem = ({field, sort:[sort], ...props}) => 
<div className="GridHeaderItem" {...props}>
    {field.name in sort && sort[field.name] === -1 && '^ '}
    {field.name}
    {field.name in sort && sort[field.name] === 1 && ' v'}
</div>



const GridHeader = ({fields, sort, onSort}) => 
<div className="GridHeader">
    {fields.map(field => <GridHeaderItem field={field} sort={sort} onClick={() => onSort(field.name)}/>)}
</div>

const VirtualScroll = ({gridHeight, count, rowHeight, row=Row, onScroll, records, skip}) => { 
    //const [records, setRecords] = useState([])
    const limit = gridHeight/rowHeight
    const Row = row


    const timeout = useRef(0)

    return (
    <div className='GridViewport' 
         onScroll={e => {
             let skip = Math.ceil(e.target.scrollTop/rowHeight)
             if (skip < 0) skip = 0
             if (skip > count - limit) skip = count - limit

             clearInterval(timeout.current)
             timeout.current = setTimeout(() => {
                 console.log(skip, limit)
                 onScroll(skip,limit)
             }, 1000)
         }} 
         style={{maxHeight: gridHeight, height: gridHeight}} >
        <div className='GridContent' 
             style={{height: count*rowHeight, minHeight: count*rowHeight, maxHeight: count*rowHeight}} >
            <div style={{height: skip*rowHeight}} />
            {records && records.map((record,i) => 
                                            <Row key={i}>
                                                {record.constructor.fields.map(field => 
                                                    <Cell>
                                                        {record[field.name] === 'object' ? record[field.name].toString() :  record[field.name]}
                                                    </Cell>)}
                                            </Row>)}
            <div style={{height: (count - skip - records.length)*rowHeight}} />
        </div>
    </div>
    )
}



const Grid = ({sort, scroll, gridHeight, overload, setScroll, count, rowHeight, records, skip, model, cellDoubleClick}) =>  {
    const viewport = useRef()
    return (
    <div className='GridViewport' 
         onScroll={e => {
             console.log((Math.abs(e.target.scrollTop - scroll)))
                if (Math.abs(e.target.scrollTop - scroll) > gridHeight*overload/5){  
                    setScroll(e.target.scrollTop); 
                    //setRecords(null)
                } }} 
         style={{maxHeight: gridHeight, height: gridHeight}}
         ref={e => e && e.scrollTo(0,scroll)}>

        <div className='GridContent' 
             style={{height: count*rowHeight, minHeight: count*rowHeight, maxHeight: count*rowHeight}} >
            <div style={{height: skip*rowHeight}} />
            {records && records.map((record,i) => 
                                            <Row key={record._id || record.key}>
                                                {model.fields.map(field => 
                                                    <Cell onDoubleClick={e => cellDoubleClick(field.name, record[field.name], record._id)}>
                                                        {record[field.name] === 'object' ? record[field.name].toString() :  record[field.name]}
                                                    </Cell>)}
                                            </Row>)}
            <div style={{height: (count - skip - records.length)*rowHeight}} />
        </div>
    </div>
    )
}


const ModelView = ({model, components:Components={Search, Count, GridHeader, Grid}, rowHeight=50, gridHeight=500, overload=2}) => {



    const [records, setRecords] = useState([])


    const [search, setSearch] = useState("")
    const [count, setCount]   = useState(0)

    const [query,  setQuery]  = useState({})
    const [cursorCalls,  setCursorCalls]  = useState({sort:[{}], skip: [0], limit: [gridHeight/rowHeight]})

    const skip = cursorCalls.skip[0]

    console.log(cursorCalls)

    //useEffect(() => {
        //setCursorCalls({sort:cursorCalls.sort, skip: skip ? [skip] : undefined, limit: overloadedRowCount ? [overloadedRowCount] : undefined})
    //}, [scroll])

    const timeout = useRef(0)
    useEffect(() => {
        clearInterval(timeout.current)
        timeout.current = setTimeout(() => {
            setQuery(searchQueryBuilder(search, model))
        }, 1000)
    },[search, model])

    useEffect(() => {
        model.count(query, cursorCalls).then(count => setCount(count))
        model.find(query, cursorCalls).then(records => Promise.all(records)).then(records => setRecords(records))
    }, [query, model, cursorCalls])

    //console.log(records)

    return (
        <>
                <Components.Search value={search} onChange={({target: {value}}) => setSearch(value)}/>
                <Components.Count> 
                    {count}
                </Components.Count> 
                <GridHeader fields={model.fields} 
                            sort={cursorCalls.sort} 
                            onSort={sort => setCursorCalls({...cursorCalls,
                                sort: [{[sort]: cursorCalls.sort[0][sort] === 1 ? -1 : 1}]
                            })}/>

                {/*records && <Grid 
                                scroll={scroll}
                                setScroll={setScroll}
                                gridHeight={gridHeight}
                                overload={overload}
                                count={count}
                                rowHeight={rowHeight}
                                records={records}
                                skip={skip}
                                model={model}
                                /> */}
                {records && <VirtualScroll 
                                skip={skip}
                                count={count}
                                records={records}
                                gridHeight={500} 
                                rowHeight={50} 
                                onScroll={(skip, limit) =>  {
                                    model.find(query, {...cursorCalls, limit: [limit], skip: [skip]}).then(records => Promise.all(records)).then(records => setRecords(records)).then(() =>
                                    setCursorCalls({...cursorCalls, limit: [limit], skip: [skip]}))
                                }}/> }
        </>
    )
}

const Admin = ({models, components:Components={ModelList, Search}}) => {
    const [selected, setSelected] = useState()


    return (
        <>
            <Components.ModelList models={models} onChange={(name) => setSelected(name)} selected={selected}/>
            <content>

                {selected && <ModelView model={models[selected]} /> }
            </content>
        </>
    )
}


function App() {
    let [models, setModels] = useState()
    let [queryText,  setQueryText]  = useState('{manufacturerName: "TOYOTA"}')
    let [query,  setQuery]  = useState({})
    models || createModels2(gql).then(models => setModels(models))

    console.log(query)
    const classes = models
    console.log(classes)
    //if (classes && Object.keys(classes).length){
        //classes.Good.find().then(goods => console.log(goods))
        //classes.Image.find().then(images => console.log(images))
    //}

    return (
        <div className="App">
            {models && <Admin models={models} />}
            {/*models && <ModelGrid model={models.Image} rowHeight={50} gridHeight={700} overload={5} query={query} 
                        cellDoubleClick={(name, text, _id) => setQueryText(`{${name}:\`${text}\`}`)}/> */}
            <textarea onChange={e => setQueryText(e.target.value)} value={queryText} />
            <button onClick={() => setQuery(eval(`(${queryText})`))}>Run</button>
        </div>
    );
}

export default App;
